from __future__ import absolute_import
from __future__ import division
import unittest
import pandas as pd
from collections import Counter
from mock import MagicMock
import pyspark.sql
from pyspark.sql import Row
from affirm.model_interpretation.shparkley.spark_shapley import (
    compute_shapley_score,
    compute_shapley_for_sample,
    ShparkleyModel
)


class TestShparkleyModel(ShparkleyModel):

    def get_required_features(self):
        return self._model.get_required_features()

    def predict(self, feature_matrix):
        pd_df = pd.DataFrame.from_dict(feature_matrix)
        preds = self._model.predict(pd_df)
        return preds


def model_predict_side_effect_function(df):
    df['score'] = (df['f1'] * 3 + df['f2'] * 5)
    return df['score'].values


class SparkShapleyTest(unittest.TestCase):

    def setUp(self):
        self.m_model = MagicMock()
        self.m_model.get_required_features.return_value = ['f1', 'f2']
        self.m_model.predict.side_effect = model_predict_side_effect_function
        self.row1 = {
            'f1': 0.01,
            'f2': 0.05,
        }
        self.row2 = {
            'f1': 0.2,
            'f2': 0.5,
        }
        self.row3 = {
            'f1': 1.0,
            'f2': 0.5,
        }
        self.row_investigate = {
            'f1': 7.0,
            'f2': 6.5,
        }
        self.m_shparkley_model = TestShparkleyModel(self.m_model)
        builder = pyspark.sql.SparkSession.builder.master('local[1]')
        self.spark = builder.appName('unittest').getOrCreate()

    def tearDown(self):
        if self.spark is not None:
            self.spark.stop()

    def test_compute_shapley_for_sample(self):
        dataset = self.spark.createDataFrame([self.row1, self.row2, self.row3])
        shapley_scores = compute_shapley_for_sample(
            df=dataset,
            model=self.m_shparkley_model,
            row_to_investigate=Row(**self.row_investigate)
        )
        sorted_shapley_scores = sorted([(k, v) for k, v in shapley_scores.items()])
        self.assertEquals(sorted_shapley_scores, [('f1', 19.79), ('f2', 30.75)])

    def test_compute_shapley_for_sample_weighted(self):

        # Add weights: [2, 1, 1]
        self.row1.update({'weight': 2})
        for row in (self.row2, self.row3):
            row.update({'weight': 1})

        dataset = self.spark.createDataFrame([self.row1, self.row2, self.row3])
        shapley_scores = compute_shapley_for_sample(
            df=dataset,
            model=self.m_shparkley_model,
            row_to_investigate=Row(**self.row_investigate),
            weight_col_name='weight',
        )

        sorted_shapley_scores = sorted([(k, v) for k, v in shapley_scores.items()])
        # row 1, which has small feature vals (=> smaller prediction) weighted more heavily means that compared to
        # baseline, the relatively large features in the sample of interest (=> bigger prediction => more different
        # from row 1) will increase both shapley values to the below compared to unweighted (19.79, 30.75)
        self.assertEquals(sorted_shapley_scores, [('f1', 20.085), ('f2', 31.125)])

    def test_compute_shapley_score(self):
        row_samples = [Row(**self.row1), Row(**self.row2), Row(**self.row3)]
        scores = compute_shapley_score(
            partition_index=1,
            rand_rows=row_samples,
            row_to_investigate=Row(**self.row_investigate),
            model=self.m_shparkley_model
        )
        expected_result = [
            ('f1', 18.0, 1.0),
            ('f1', 20.4, 1.0),
            ('f1', 20.97, 1.0),
            ('f2', 30.0, 1.0),
            ('f2', 30.0, 1.0),
            ('f2', 32.25, 1.0)
        ]
        self.assertEqual(sorted(list(scores)), expected_result)

    def test_compute_shapley_score_weighted(self):

        # Get unweighted result for reference
        row_samples = [Row(**self.row1), Row(**self.row2), Row(**self.row3)]
        unweighted_result = compute_shapley_score(
            partition_index=1,
            rand_rows=row_samples,
            row_to_investigate=Row(**self.row_investigate),
            model=self.m_shparkley_model
        )

        unweighted_result = sorted(list(unweighted_result))

        # Add weights: [2, 1, 1]
        self.row1.update({'weight': 2})
        for row in (self.row2, self.row3):
            row.update({'weight': 1})
        row_samples__weighted_first_double = [Row(**self.row1), Row(**self.row2), Row(**self.row3)]

        weighted_first_double_result = compute_shapley_score(
            partition_index=1,
            rand_rows=row_samples__weighted_first_double,
            row_to_investigate=Row(**self.row_investigate),
            model=self.m_shparkley_model,
            weight_col_name='weight',
        )

        weights_found = Counter()
        for (ft_name_uw, ft_val_uw, ft_val_weight_uw), (ft_name_double, ft_val_double, ft_val_weight) in zip(
                unweighted_result,
                sorted(list(weighted_first_double_result))
        ):
            self.assertEqual(ft_name_uw, ft_name_double)
            self.assertEqual(ft_val_uw, ft_val_double)
            weights_found[ft_val_weight] += 1

        # Four (two rows times two features) of \phi_j^ms of weight 1, two (one rwo times two features) of
        # weight 2
        self.assertDictEqual(weights_found, {1.0: 4, 2.0: 2})

    def test_efficiency_property(self):
        # The Shapley value must satisfy the Efficiency property
        # The feature contributions must add up to the difference of prediction for x and the average.
        dataset = self.spark.createDataFrame([self.row1, self.row2, self.row3])
        shapley_scores = compute_shapley_for_sample(
            df=dataset,
            model=self.m_shparkley_model,
            row_to_investigate=Row(**self.row_investigate)
        )
        total_shapley_value = sum([v for _, v in shapley_scores.items()])
        predicted_value_for_row = model_predict_side_effect_function(
            pd.DataFrame.from_dict([self.row_investigate])
        )
        rows = [self.row1, self.row2, self.row3]
        scores = model_predict_side_effect_function(pd.DataFrame.from_dict(rows))
        mean_prediction_on_dataset = sum(scores)/len(rows)
        self.assertAlmostEquals(
            first=total_shapley_value,
            second=predicted_value_for_row - mean_prediction_on_dataset,
            delta=0.01
        )



from __future__ import absolute_import
from __future__ import division
from typing import TYPE_CHECKING
import numpy as np
from abc import ABCMeta, abstractmethod
from six import add_metaclass
from pyspark.sql import functions as fn
from itertools import permutations, cycle, chain
from future.builtins import zip, range

if TYPE_CHECKING:
    from pyspark.sql import Row, DataFrame
    from typing import List, Iterable, Tuple, Dict, Set, Generator, Optional, Any


@add_metaclass(ABCMeta)
class ShparkleyModel(object):
    """
    Abstract class for computing Shapley values.
    """
    def __init__(self, model):
        # type: (Any) -> None
        self._model = model

    @abstractmethod
    def get_required_features(self):
        # type: () -> Set[str]
        """
        Returns the set of feature names
        :return: Set of feature names
        """
        raise NotImplementedError

    @abstractmethod
    def predict(self, feature_matrix):
        # type: (List[Dict[str, Any]]) -> List[float]
        """
        Run the machine learning model on a feature matrix and return the predictions for each row.
        :param feature_matrix: Row of feature vectors. Each entry is a dictionary mapping
        from the feature name to feature value
        :return: Model predictions for all feature vectors
        """
        raise NotImplementedError


def compute_shapley_score(partition_index, rand_rows, row_to_investigate, model, weight_col_name=None):
    # type: (int, Iterable[Row], Row, ShparkleyModel, Optional[str]) -> Generator[Tuple[str, float, float], None, None]
    """
    Computes the shapley marginal contribution for each feature in the feature vector over all
    samples in the partition.
    The algorithm is based on a monte-carlo approximation:
    https://christophm.github.io/interpretable-ml-book/shapley.html#fn42
    :param partition_index: Index of spark partition which will serve as a seed to numpy
    :param rand_rows: Sampled rows of the dataset in the partition
    :param row_to_investigate: Feature vector for which we need to compute shapley scores
    :param model: ShparkleyModel object which implements the predict function.
    :param weight_col_name: column name with row weights to use when sampling the training set
    :return: Generator of tuple of feature and shapley marginal contribution
    """
    required_features = list(model.get_required_features())
    random_feature_permutation = np.random.RandomState(partition_index).permutation(required_features)
    # We cycle through permutations in cases where the number of samples is more than
    # the number of features
    permutation_iter = cycle(permutations(random_feature_permutation))
    feature_vector_rows = []
    rand_row_weights = []
    for rand_row in rand_rows:  # take sample z from training set
        rand_row_weights.append(rand_row[weight_col_name] if weight_col_name is not None else 1)
        feature_permutation = next(permutation_iter)  # choose permutation o
        # gather: {z_1, ..., z_p}
        feat_vec_without_feature = {feat_name: rand_row[feat_name] for feat_name in required_features}
        feat_vec_with_feature = feat_vec_without_feature.copy()
        for feat_name in feature_permutation:  # for random feature k.
            # x_+k = {x_1, ..., x_k, .. z_p}
            feat_vec_with_feature[feat_name] = row_to_investigate[feat_name]
            # x_-k = {x_1, ..., x_{k-1}, z_k, ..., z_p}
            # store (x_+k, x_-k)
            feature_vector_rows.append(feat_vec_with_feature.copy())
            feature_vector_rows.append(feat_vec_without_feature.copy())
            # (x_-k = x_+k)
            feat_vec_without_feature[feat_name] = row_to_investigate[feat_name]

    if len(feature_vector_rows) == 0:
        return
    preds = model.predict(feature_vector_rows)
    feature_iterator = chain.from_iterable(cycle(permutations(random_feature_permutation)))
    for pred_index, feature in zip((range(0, len(preds), 2)), feature_iterator):
        marginal_contribution = preds[pred_index] - preds[pred_index + 1]

        #  There is one weight added per random row visited.
        #  For each random row visit, we generate 2 predictions for each required feature.
        #  Therefore, to get index into rand_row_weights, we need to divide
        #  prediction index by 2 * number of features, and take the floor of this.
        weight = rand_row_weights[pred_index // (len(required_features) * 2)]

        yield (str(feature), float(marginal_contribution), float(weight))


def compute_shapley_for_sample(df, model, row_to_investigate, weight_col_name=None):
    # type: (DataFrame, ShparkleyModel, Row, Optional[str]) -> Dict[str, float]
    """
    Compute shapley values for all features in a given feature vector of interest.

    :param df: Training dataset
    :param model: ShparkleyModel object which implements the predict function.
    :param row_to_investigate: Feature vector for which we need to compute shapley scores
    :param weight_col_name: column name with row weights to use when sampling the training set
    :return: Dictionary of feature mapping to its corresponding shapley value.
    """
    shapley_df = (
        df.rdd.mapPartitionsWithIndex(
            lambda idx, rows: compute_shapley_score(idx, rows, row_to_investigate, model, weight_col_name),
            preservesPartitioning=True,
        )
    ).toDF(["feature", "marginal_contribution", "weight"])

    return dict(
        shapley_df.groupBy("feature")
                  .agg((
                                 fn.sum(shapley_df.marginal_contribution * shapley_df.weight) /
                                 fn.sum(shapley_df.weight)
                    ).alias("shapley_value"))
                  .collect()
    )



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

REQUIRED_PACKAGES = [
    'future~=0.16',
    'mock~=2.0.0',
    'pyspark>=2.4.3',
]

setuptools.setup(
    name="shparkley",
    version="0.0.2",
    install_requires=REQUIRED_PACKAGES,
    author="niloygupta",
    author_email="niloy.gupta@affirm.com",
    description="Scaling Shapley Value computation using Spark",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/Affirm/shparkley",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
